from forest.common import proc_utils
import shutil

class GitTools:

    def __init__(self, srcdir) -> None:
        self.srcdir = srcdir

    def clone(self, 
            server: str, 
            repository: str, 
            tag: str,
            proto='ssh', 
            recursive=False,
            depth=None):

        if proto == 'ssh':
            addr = f'git@{server}:{repository}'
        elif proto == 'https':
            addr = f'https://{server}/{repository}'
        else:
            # TODO more specific exception
            raise ValueError(f'unsupported protocol "{proto}"')
        
        # create command
        cmd = ['git', 'clone', '--branch', tag]
        
        if recursive:
            cmd.append('--recursive')

        if depth is not None:
            cmd.extend(['--depth', depth])

        cmd.extend([addr, self.srcdir])

        # clone, and delete the source folder on failure
        # (either exception or git returns != 0) 
        try:
             clone_ok = proc_utils.call_process(args=cmd)
             if not clone_ok:
                 self.rm()
        except BaseException as e:
            # remove src and re-raise exception
            self.rm()
            raise e

        return True

    def checkout(self, tag):
        return proc_utils.call_process(['git', 'checkout', tag], cwd=self.srcdir)

    def rm(self):
        shutil.rmtree(self.srcdir,  ignore_errors=True)




from distutils.command.build import build
import os 
from tempfile import TemporaryDirectory

from forest.cmake_tools import CmakeTools
from forest.common import eval_handler
from forest.common import print_utils
from forest.common.fetch_handler import CustomFetcher
from forest.common.print_utils import ProgressReporter
from forest.common import proc_utils

class BuildHandler:

    # entries in this cache have already been built
    build_cache = set()

    def __init__(self, pkgname) -> None:
        
        # pkgname
        self.pkgname = pkgname

        # print function to report progress
        self.pprint = ProgressReporter.get_print_fn(pkgname)

        # pre and post build commands
        self.pre_build_cmd = list()
        self.post_build_cmd = list()

        # env hook to install
        self.env_hook_content = None

    
    def pre_build(self, builddir):
        for cmd in self.pre_build_cmd:
            proc_utils.call_process(args=[cmd], cwd=builddir, shell=True)

    def post_build(self, builddir):
        for cmd in self.post_build_cmd:
            proc_utils.call_process(args=[cmd], cwd=builddir, shell=True)
    
    def install_env_hook(self, installdir):
        if self.env_hook_content is None:
            return 

        os.makedirs(f'{installdir}/share/forest_env_hook', exist_ok=True)

        with open(f'{installdir}/share/forest_env_hook/{self.pkgname}.bash', 'w') as f:
            f.write(self.env_hook_content)

        self.pprint('installed environment hook, re-source your setup.bash')

    def build(self, 
              srcdir: str, 
              builddir: str, 
              installdir: str,
              buildtype: str,
              jobs: int,
              reconfigure=False) -> bool:
        """
        Carry out the build procedure for this package.

        Args:
            srcdir (str): directory with source code as cloned by the FetchHandler
            builddir (str): directory where build artifacts must be generated
            installdir (str): directory where the build output will be copied
            buildtype (str): a string describing the build type (e.g., Release)
            jobs (int): number of parallel jobs to use
            reconfigure (bool, optional): flag indicating if configuration step must be forces (cmake specific). Defaults to False.
        """
        
        if self.pkgname in BuildHandler.build_cache:
            return True 

        BuildHandler.build_cache.add(self.pkgname)
        self.pprint('no build action required')
        return True 

    
    @classmethod
    def from_yaml(cls, pkgname, data, recipe):
        buildtype = data['type']
        builder = None
        
        if buildtype == 'cmake':
            builder = CmakeBuilder.from_yaml(pkgname=pkgname, data=data)
        elif buildtype == 'custom':
            builder = CustomBuilder.from_yaml(pkgname=pkgname, data=data)
        else: 
            raise ValueError(f'unsupported build type "{buildtype}"')
        
        # pre and post-build (unconditional)
        pre_build = recipe.get('pre_build', list())
        post_build = recipe.get('post_build', list())

        # conditional
        eh = eval_handler.EvalHandler.instance()

        pre_build_if = recipe.get('pre_build_if', dict())
        pre_build.extend(eh.parse_conditional_dict(pre_build_if))
        pre_build = map(eh.process_string, pre_build)

        post_build_if = recipe.get('post_build_if', dict())
        post_build.extend(eh.parse_conditional_dict(post_build_if))
        post_build = map(eh.process_string, post_build)

        # apply
        builder.pre_build_cmd = list(pre_build)
        builder.post_build_cmd = list(post_build)

        print('*****')
        print(builder.pre_build_cmd)

        # env hooks
        env_hooks = data.get('env_hooks', list())
        env_hook_content = str()
        for hline in env_hooks:
            env_hook_content += eh.process_string(hline, shell=False)
            env_hook_content += '\n'
        
        builder.env_hook_content = env_hook_content if len(env_hooks) > 0 else None


        return builder


class CustomBuilder(BuildHandler):

    def __init__(self, pkgname) -> None:
        super().__init__(pkgname)
        self.commands = list()

    def build(self, srcdir: str, builddir: str, installdir: str, buildtype: str, jobs: int, reconfigure=False) -> bool:
        
        # evaluator
        eh = eval_handler.EvalHandler.instance()

        # check if package in cache
        if self.pkgname in BuildHandler.build_cache:
            self.pprint('already built, skipping')
            return True
        
        # create source folder
        if not os.path.exists(srcdir):
            os.mkdir(srcdir)

        self.pprint('building...')
        with TemporaryDirectory(prefix="foresttmp-") as tmpdir:
            for cmd in self.commands:
                cmd_p = eh.process_string(cmd, {'srcdir': srcdir, 'installdir': installdir, 'jobs': jobs})
                if not proc_utils.call_process(cmd_p, cwd=tmpdir, shell=True, print_on_error=True):
                    self.pprint(f'{cmd_p} failed')
                    return False 

        # install hooks 
        self.install_env_hook(installdir)

        # save to cache and exit
        BuildHandler.build_cache.add(self.pkgname)
        
        return True 

    @classmethod
    def from_yaml(cls, pkgname, data):
        ret = CustomBuilder(pkgname=pkgname)
        ret.commands = list(data['cmd'])
        return ret

class CmakeBuilder(BuildHandler):

    def __init__(self, pkgname, cmake_args=None, cmakelists='.', target='install') -> None:

        super().__init__(pkgname=pkgname)
        self.cmake_args = cmake_args if cmake_args is not None else list()
        self.cmakelists_folder = cmakelists
        self.target = target
    
    
    @classmethod
    def from_yaml(cls, pkgname, data):

        # check if we must skip this build,
        # and return a dummy builder if so
        eh = eval_handler.EvalHandler.instance()
        skip_condition = data.get('skip_if', 'False')
        if eh.eval_condition(skip_condition):
            return BuildHandler(pkgname=pkgname)

        # first, parse cmake arguments
        args = data.get('args', list())
        args_if = data.get('args_if', dict())

        # parse conditional cmake arguments
        args.extend(eh.parse_conditional_dict(args_if))

        # process all args through the shell
        args = map(eh.process_string, args)

        return CmakeBuilder(pkgname=pkgname, 
                            cmake_args=args,
                            cmakelists=data.get('cmakelists', '.'),
                            target=data.get('target', 'install'))


    def build(self, 
              srcdir: str, 
              builddir: str, 
              installdir: str,
              buildtype: str,
              jobs: int,
              reconfigure=False) -> bool:

        # check if package in cache
        if self.pkgname in BuildHandler.build_cache:
            self.pprint('already built, skipping')
            return True

        # path to folder containing cmakelists
        cmakelists = os.path.join(srcdir, self.cmakelists_folder)
        
        # create build folder if needed
        if not os.path.exists(builddir):
            os.mkdir(builddir)

        # create cmake tools
        cmake = CmakeTools(srcdir=cmakelists, builddir=builddir)

        # configure
        if not cmake.is_configured() or reconfigure:
            # set install prefix and build type (only on first or forced configuration)
            cmake_args = list()
            cmake_args.append(f'-DCMAKE_INSTALL_PREFIX={installdir}')
            cmake_args.append(f'-DCMAKE_BUILD_TYPE={buildtype}')
            cmake_args += self.cmake_args  # note: flags from recipes as last entries to allow override

            self.pprint('running cmake...')
            if not cmake.configure(args=cmake_args):
                self.pprint('configuring failed')
                return False

        # pre-build
        self.pre_build(builddir)

        # build
        self.pprint('building...')
        if not cmake.build(target=self.target, jobs=jobs):
            self.pprint('build failed')
            return False 

        # post-build
        self.post_build(builddir)

        # install hooks 
        self.install_env_hook(installdir)
        
        # save to cache and exit
        BuildHandler.build_cache.add(self.pkgname)

        return True




import os
from forest.common import proc_utils
from forest.common import config_handler
import inspect
from typing import List

class EvalHandler:

    _instance = None

    # modes are strings that are used to conditionally add cmake args
    # (and possibly other things, too!)
    modes = set()

    class Locals:

        def __init__(self) -> None:
            self.ubuntu_release = float(proc_utils.get_output('lsb_release -rs'.split(' ')))
            self.shell = EvalHandler.Locals.shell
            self.env = EvalHandler.Locals.env
            self.config = config_handler.ConfigHandler.instance()
            self.mode = EvalHandler.Locals.mode

        @staticmethod
        def mode(mode: str) -> bool:
            return mode in EvalHandler.modes

        @staticmethod
        def shell(cmd: str) -> str:
            ret = proc_utils.get_output(args=[cmd], shell=True)
            if ret is None:
                return ''
            else:
                return ret

        @staticmethod
        def env(key: str) -> str:
            return os.environ.get(key, '')
    
    
    def __init__(self) -> None:
        self.locals = EvalHandler.Locals()

    
    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    
    @classmethod
    def echo(cls, text: str) -> str:
        return proc_utils.get_output(args=['/bin/echo "' + text + '"'], shell=True)
        
    
    @classmethod
    def print_available_locals(cls):

        print('{:20s} {}'.format('name', 'type'))
        print('{:20s} {}'.format('----', '----'))
        
        for k, v in cls.instance().locals.__dict__.items():

            if callable(v):
                descr = f'function{inspect.signature(v)}'
            else:
                descr = type(v).__name__
            
            print('{:20s} {}'.format(k, descr))

    def eval_string(self, code: str, ret_type=str, default=None, throw_on_failure=True):
        try:
            ret = ret_type(eval(code, None, self.locals.__dict__))
            if proc_utils.call_process_verbose:
                print(f'{code} evaluated to {ret}')
            return ret
        except BaseException as e:
            if proc_utils.call_process_verbose:
                print(f'failed to evaluate "{code}" to {ret_type} with error "{str(e)}"')
            if throw_on_failure:
                raise e
            return default

    def eval_condition(self, code: str):
        return self.eval_string(code=code, ret_type=bool, default=False, throw_on_failure=False)

    def format_string(self, text: str, locals=None) -> str:
        if locals is None:
            locals = dict()
        ret = text.format(**self.locals.__dict__, **locals)
        if proc_utils.call_process_verbose:
            print(f'formatted string "{text}" into "{ret}"')
        return ret


    def parse_conditional_dict(self, args_if) -> List[str]:

        args = list()

        # parse conditional cmake arguments
        eh = EvalHandler.instance()
        for k, v in args_if.items():

            add_arg = False

            # process key through the shell
            k = self.process_string(k)
            
            # check if key is an active mode, 
            # or is an expression returning True
            if k in EvalHandler.modes:
                add_arg = True
            elif eh.eval_condition(code=k):
                add_arg = True
            else:
                add_arg = False

            if not add_arg:
                continue

            # extend args with all conditional args
            if not isinstance(v, list):
                v = [v]
            
            args.extend(v)

        print(args)
        
        return args

        
    def process_string(self, text, locals=None, shell=True):
        # check if text is in the form ${code}
        is_expression = len(text) >= 3 and text[0:2] == '${' and text[-1] == '}'
        if is_expression:
            return self.eval_string(text[2:-1])
        
        # otherwise echo through the shell and then apply format rules
        ret = str(text)

        if shell:
            ret = EvalHandler.echo(ret)

        ret = self.format_string(ret, locals)
        
        return ret



from typing import List

class ConfigHandler:

    _instance = None

    @classmethod
    def instance(cls) -> 'ConfigHandler':
        if cls._instance is None:
            cls._instance = ConfigHandler()
        return cls._instance

    def set_config_variables(self, var_list: List[str]):
        for var_str in var_list:
            k, v = var_str.split(':=')
            self.__dict__[k] = v
            
        


import os

from forest.cmake_tools import CmakeTools
from . import package
from .print_utils import ProgressReporter

_build_cache = dict()

def build_package(pkg: package.Package, 
                  srcroot: str, 
                  buildroot: str, 
                  installdir: str,
                  buildtype: str,
                  jobs: int,
                  reconfigure=False):

    # source dir and build dir
    srcdir = os.path.join(srcroot, pkg.name)
    builddir = os.path.join(buildroot, pkg.name)

    # doit!
    return pkg.builder.build(srcdir=srcdir, builddir=builddir, installdir=installdir, 
                      buildtype=buildtype, jobs=jobs, reconfigure=reconfigure)



# function to install one package with dependencies
@ProgressReporter.count_calls
def install_package(pkg: str,
                    srcroot: str,
                    buildroot: str,
                    installdir: str,
                    buildtype: str,
                    jobs: int,
                    reconfigure=False, 
                    no_deps=False):
    
    """
    Fetch a recipe file from the default path using the given package name, 
    and perform the required cloning and building steps of the package and 
    its dependencies

    Returns:
        bool: success flag
    """

    # custom print
    pprint = ProgressReporter.get_print_fn(pkg)

    # retrieve package info from recipe
    try:
        pkg = package.Package.from_name(name=pkg)
    except FileNotFoundError:
        pprint(f'recipe file not found (searched in {package.Package.get_recipe_path()})')
        return False

    # install dependencies if not found
    for dep in pkg.depends:

        # this dependency build directory name (if exists)
        dep_builddir = os.path.join(buildroot, dep)

        # if dependency is built by this ws, trigger build
        if os.path.exists(dep_builddir):
            
            # dependency found and built by forest -> trigger build
            pprint(f'depends on {dep} -> build found, building..')   

            ok = install_package(dep, srcroot, buildroot, installdir, 
                    buildtype, jobs, reconfigure, no_deps=True)   

            if not ok:
                pprint(f'failed to build dependency {dep}')
                return False 

            # go to next dependency
            continue

        # if no-deps mode, skip dependency installation
        if no_deps:
            pprint('skipping dependencies')
            continue
        
        # try to find-package this dependency
        dep_found = CmakeTools.find_package(dep)

        if not dep_found:
            # dependency not found -> install it
            pprint(f'depends on {dep} -> not found, installing..')
            
            # note: reconfigure needed if there's build but not install
            ok = install_package(dep, srcroot, buildroot, installdir, 
                    buildtype, jobs, reconfigure)   

            if not ok:
                pprint(f'failed to install dependency {dep}')
                return False

        else:
            # dependency found and not built by forest -> nothing to do
            pprint(f'depends on {dep} -> found')
    
    # use the fetcher! (if not build only)
    srcdir = os.path.join(srcroot, pkg.name)
    if not pkg.fetcher.fetch(srcdir):
        pprint('failed to fetch package')
        return False 
    
    # configure and build
    ok = build_package(pkg=pkg, 
                       srcroot=srcroot, 
                       buildroot=buildroot, 
                       installdir=installdir, 
                       buildtype=buildtype, 
                       jobs=jobs, 
                       reconfigure=reconfigure)

    if ok:
        pprint('ok')

    return ok


def write_setup_file(srcdir, installdir):
    
    """
    Write a setup file to the root directory.
    """

    this_dir = os.path.dirname(os.path.abspath(__file__))
    setup_template = os.path.join(this_dir, 'setup.bash')
    with open(setup_template, 'r') as f:
        content = f.read()
        content = content.replace('£PREFIX£', os.path.realpath(installdir))
        content = content.replace('£SRCDIR£', os.path.realpath(srcdir))
    
    setup_file = os.path.join(installdir, '..', 'setup.bash')
    if not os.path.exists(setup_file):
        with open(setup_file, 'w') as f:
            f.write(content)


def check_ws_file(rootdir):
    
    """
    Check forest marker file exists
    """

    ws_file = os.path.join(rootdir, '.forest')

    return os.path.exists(ws_file)


def write_ws_file(rootdir):
    
    """
    Write a hidden file to mark the forest root directory
    """
    
    ws_file = os.path.join(rootdir, '.forest')

    if check_ws_file(rootdir=rootdir):
        print('workspace already initialized')
        return False

    with open(ws_file, 'w') as f:
        f.write('# forest marker file \n')
        return True




import subprocess
import sys 
from forest.common import print_utils

call_process_verbose = False

def call_process(args, cwd='.', input=None, verbose=False, print_on_error=True, shell=False) -> bool:

    # convert args to string
    args = list(map(str, args))

    if verbose or call_process_verbose:
        if shell:
            print(f'calling shell with command "{args}"')
        else:
            print('calling "{}"'.format(' '.join(args)))

    executable = '/bin/bash' if shell else None

    if call_process_verbose or verbose:
        # run will print output to terminal
        proc = subprocess.run(args=args, 
                        cwd=cwd, 
                        input=input, 
                        shell=shell, 
                        executable=executable)

        print(f'returned {proc.returncode}')

        return proc.returncode == 0 

    try:
        # run with output/error redirection and exit status check
        pr = subprocess.run(args=args, 
                        stdout=subprocess.PIPE, 
                        stderr=subprocess.STDOUT, 
                        cwd=cwd, 
                        input=input, 
                        shell=shell, 
                        executable=executable,
                        check=True)

        print_utils.log_file.write(pr.stdout.decode())

    except subprocess.CalledProcessError as e:
        # on error, print output (includes stderr)
        print_utils.log_file.write(e.output.decode())

        if print_on_error and not verbose:
            print(e.output.decode(), file=sys.stderr)
            
        return False

    return True


def get_output(args, cwd='.', input=None, verbose=False, print_on_error=True, shell=False):

    if verbose or call_process_verbose:
        if shell:
            print(f'calling shell with command "{args}"')
        else:
            print('calling "{}"'.format(' '.join(args)))

    try:
        # check_output will not print
        out = subprocess.check_output(args=args, cwd=cwd, input=input, shell=shell)
        ret = out.decode().strip()
        if verbose or call_process_verbose:
            print('calling "{}" returned "{}"'.format(' '.join(args), ret))
        return ret
    except subprocess.CalledProcessError as e:
        # on error, print output and errors
        if print_on_error:
            print('stdout: ' + e.output.decode(), file=sys.stderr)
            print('stderr: ' + e.stderr.decode(), file=sys.stderr)
        return None



import os 
import shutil
import getpass
from tempfile import TemporaryDirectory

from forest.git_tools import GitTools
from forest.common import proc_utils
from forest.common.print_utils import ProgressReporter
from forest.common.eval_handler import EvalHandler

class FetchHandler:
    
    """
    Abstract interface to a fetch handler, i.e., a class that incorporates
    the logic behind the fetching of a package's source code.
    As an exception, we also consider this base class to describe the process
    of fetching binary distributions such as debian packages (even though in
    this case sources are not involved).
    
    Concrete instances of this class are created via the from_yaml() factory method.
    """

    def __init__(self, pkgname) -> None:
        """
        Construct the FetchHandler

        Args:
            pkgname (str): name of the package
        """
        self.pkgname = pkgname

        # print function to report progress
        self.pprint = ProgressReporter.get_print_fn(pkgname)

    
    def fetch(self, srcdir):
        """
        Carry out the fetch operation on the package.

        Args:
            srcdir (str): directory where sources are cloned/copied
        """
        self.pprint('no fetch action required')
        return True 

    @classmethod
    def from_yaml(cls, pkgname, data, recipe) -> 'FetchHandler':
        """ 
        Factory method to instantiate concrete fetchers given their
        yaml description.

        Args:
            pkgname (str): name of the package
            data (dict): the 'clone' entry of the yaml recipe, parsed 
            into a python dictionary

        Raises:
            ValueError: the specified fetcher type is not supported

        Returns:
            FetchHandler: the requested object
        """

        fetchtype = data['type']
        if fetchtype == 'git':
            return GitFetcher.from_yaml(pkgname=pkgname, data=data)
        elif fetchtype == 'deb':
            return DebFetcher.from_yaml(pkgname=pkgname, data=data)
        elif fetchtype == 'custom':
            return CustomFetcher.from_yaml(pkgname=pkgname, data=data)
        else: 
            raise ValueError(f'unsupported fetch type "{fetchtype}"')


class CustomFetcher(FetchHandler):

    def __init__(self, pkgname) -> None:
        super().__init__(pkgname)
        self.commands = list()

    def fetch(self, srcdir):
        
        # evaluator
        eh = EvalHandler.instance()
        
        if os.path.exists(srcdir):
            self.pprint(f'source code  already exists, skipping clone')
            return True
        
        # create source folder
        os.mkdir(srcdir)

        # run commands
        with TemporaryDirectory(prefix="foresttmp-") as tmpdir:
            for cmd in self.commands:
                cmd_p = eh.process_string(cmd, {'srcdir': srcdir})
                if not proc_utils.call_process(cmd_p, cwd=tmpdir, shell=True, print_on_error=True):
                    self.pprint(f'{cmd_p} failed')
                    shutil.rmtree(srcdir, ignore_errors=True)
                    return False 
        
        return True

    @classmethod
    def from_yaml(cls, pkgname, data):
        ret = CustomFetcher(pkgname=pkgname)
        ret.commands = list(data['cmd'])
        return ret

class GitFetcher(FetchHandler):

    # set this variable to override git clone protocol (e.g., to https)
    proto_override = None

    # this sets the global clone depth
    depth_override = None
    
    def __init__(self, pkgname, server, repository, tag=None, proto='ssh', recursive=False) -> None:

        super().__init__(pkgname=pkgname)
        self.tag = tag
        self.server = server
        self.repository = repository
        self.proto = proto if self.proto_override is None else self.proto_override
        self.recursive = recursive
    
    @classmethod
    def from_yaml(cls, pkgname, data):
        
        eh = EvalHandler.instance()
        
        # process tag
        tag = data['tag']
        tag = eh.process_string(tag)

        # parse tag_if (if present)
        tag_if = data.get('tag_if', dict())
        tag_if_parsed = eh.parse_conditional_dict(tag_if)

        # tag_if has precendence
        if len(tag_if_parsed) == 1:
            tag = tag_if_parsed[0]

        if len(tag_if_parsed) > 1:
            raise RuntimeError(f'[{pkgname}] tag_if conditions must be mutually exclusive')

        return GitFetcher(pkgname=pkgname, 
                          server=data['server'],
                          repository=data['repository'],
                          tag=tag,
                          proto=data.get('proto', 'ssh'),
                          recursive=data.get('recursive', False))


    def fetch(self, srcdir) -> bool:
        
        # custom print shorthand
        pprint = self.pprint

        # create git tools
        git = GitTools(srcdir=srcdir)

        eh = EvalHandler.instance()
        tag_processed = eh.process_string(self.tag)

        # check existance
        if os.path.exists(srcdir):
            pprint(f'source code already exists, skipping clone')
            return True

        pprint(f'cloning source code ({self.proto})...')
        
        if not git.clone(server=self.server, 
                        repository=self.repository, 
                        tag=tag_processed,
                        proto=self.proto, 
                        recursive=self.recursive,
                        depth=GitFetcher.depth_override):
            pprint(f'unable to clone source code (tag {tag_processed})')
            git.rm()
            return False

        return True


class DebFetcher(FetchHandler):

    # user password
    pwd = None
    superuser = 'root'

    def __init__(self, pkgname, debname: str) -> None:
        super().__init__(pkgname)
        # note: expand environment variables between {curly braces}
        # example: 'ros-{ROS_DISTRO}-moveit-core` becomes 'ros-melodic-moveit-core'
        self.debname = debname.format(**os.environ)

    
    def fetch(self, srcdir) -> bool:

        # custom print shorthand
        pprint = self.pprint

        pkg_already_installed = proc_utils.call_process(args=['dpkg', '-s', self.debname], print_on_error=False)

        if pkg_already_installed:
            pprint(f'{self.debname} already installed')
            return True
            
        pprint(f'installing {self.debname} from apt')
        
        if getpass.getuser() != DebFetcher.superuser and DebFetcher.pwd is None:
            pwd = getpass.getpass()
            pprint('got password!')
            DebFetcher.pwd = pwd

        return proc_utils.call_process(args=['sudo', '-Sk', 'apt', 'install', '-y', self.debname], 
                                       input=(DebFetcher.pwd + '\n').encode())

    
    @classmethod
    def from_yaml(cls, pkgname, data):
        return DebFetcher(pkgname=pkgname, 
                          debname=data['debname'])
